import gdsfactory as gf
from gdsfactory.typings import Float2, LayerSpec

from gf180mcu.guardring import pcmpgr_gen
from gf180mcu.layers import layer
from gf180mcu.via_generator import via_stack


@gf.cell
def res(
    l_res: float = 0.1,
    w_res: float = 0.1,
    res_type: str = "rm1",
    label: bool = False,
    r0_label: str = "",
    r1_label: str = "",
) -> gf.Component:
    """Returns 2-terminal Metal resistor by specifying parameters.

    Args:
        l_res: length of resistor.
        w_res: width of resistor.
        res_type: type of resistor.
        label: label generation.
        r0_label: label for resistor.
        r1_label: label for resistor.

    """
    c = gf.Component("res_dev")

    m_ext = 0.28

    if res_type == "rm1":
        m_layer = layer["metal1"]
        res_layer = layer["metal1_res"]
        m_label_layer = layer["metal1_label"]
    elif res_type == "rm2":
        m_layer = layer["metal2"]
        res_layer = layer["metal2_res"]
        m_label_layer = layer["metal2_label"]
    elif res_type == "rm3":
        m_layer = layer["metal3"]
        res_layer = layer["metal3_res"]
        m_label_layer = layer["metal3_label"]
    else:
        m_layer = layer["metaltop"]
        res_layer = layer["metal6_res"]
        m_label_layer = layer["metaltop_label"]

    res_mk = c.add_ref(gf.components.rectangle(size=(l_res, w_res), layer=res_layer))

    m_rect = c.add_ref(
        gf.components.rectangle(size=(l_res + (2 * m_ext), w_res), layer=m_layer)
    )
    m_rect.dxmin = res_mk.dxmin - m_ext
    m_rect.dymin = res_mk.dymin

    # labels generation
    if label == 1:
        c.add_label(
            r0_label,
            position=(
                res_mk.dxmin + (res_mk.dxsize / 2),
                res_mk.dymin + (res_mk.dysize / 2),
            ),
            layer=m_label_layer,
        )
        c.add_label(
            r1_label,
            position=(
                m_rect.dxmin + (res_mk.dxmin - m_rect.dxmin) / 2,
                m_rect.dymin + (m_rect.dysize / 2),
            ),
            layer=m_label_layer,
        )

    return c


@gf.cell
def plus_res_inst(
    l_res: float = 0.1,
    w_res: float = 0.1,
    res_type: str = "nplus_s",
    sub: bool = False,
    cmp_res_ext: float = 0.1,
    con_enc: float = 0.1,
    cmp_imp_layer: LayerSpec = layer["nplus"],
    sub_imp_layer: LayerSpec = layer["pplus"],
    label: bool = False,
    r0_label: str = "",
    r1_label: str = "",
    sub_label: str = "",
) -> gf.Component:
    """Returns 2-terminal Metal resistor by specifying parameters."""
    c = gf.Component()

    sub_w: float = 0.36
    np_enc_cmp: float = 0.16
    pp_enc_cmp: float = 0.16
    comp_spacing: float = 0.72
    sab_res_ext = 0.22

    res_mk = c.add_ref(
        gf.components.rectangle(size=(l_res, w_res), layer=layer["res_mk"])
    )

    if "plus_u" in res_type:
        sab_rect = c.add_ref(
            gf.components.rectangle(
                size=(res_mk.dxsize, res_mk.dysize + (2 * sab_res_ext)),
                layer=layer["sab"],
            )
        )
        sab_rect.dxmin = res_mk.dxmin
        sab_rect.dymin = res_mk.dymin - sab_res_ext

    cmp = c.add_ref(
        gf.components.rectangle(
            size=(res_mk.dxsize + (2 * cmp_res_ext), res_mk.dysize),
            layer=layer["comp"],
        )
    )
    cmp.dxmin = res_mk.dxmin - cmp_res_ext
    cmp.dymin = res_mk.dymin

    cmp_con = via_stack(
        x_range=(cmp.dxmin, res_mk.dxmin + con_enc),
        y_range=(cmp.dymin, cmp.dymax),
        base_layer=layer["comp"],
        metal_level=1,
    )

    cmp_con_arr = c.add_ref(
        component=cmp_con,
        rows=1,
        columns=2,
        column_pitch=cmp_res_ext - con_enc + res_mk.dxsize,
    )  # comp contact array

    # labels generation
    if label == 1:
        c.add_label(
            r0_label,
            position=(
                cmp_con_arr.dxmin + (cmp_con.dxsize / 2),
                cmp_con_arr.dymin + (cmp_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )
        c.add_label(
            r1_label,
            position=(
                cmp_con_arr.dxmax - (cmp_con.dxsize / 2),
                cmp_con_arr.dymin + (cmp_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

    cmp_imp = c.add_ref(
        gf.components.rectangle(
            size=(cmp.dxsize + (2 * np_enc_cmp), cmp.dysize + (2 * np_enc_cmp)),
            layer=cmp_imp_layer,
        )
    )
    cmp_imp.dxmin = cmp.dxmin - np_enc_cmp
    cmp_imp.dymin = cmp.dymin - np_enc_cmp

    if sub == 1:
        sub_rect = c.add_ref(
            gf.components.rectangle(size=(sub_w, w_res), layer=layer["comp"])
        )
        sub_rect.dxmax = cmp.dxmin - comp_spacing
        sub_rect.dymin = cmp.dymin

        # sub_rect contact
        sub_con = c.add_ref(
            via_stack(
                x_range=(sub_rect.dxmin, sub_rect.dxmax),
                y_range=(sub_rect.dymin, sub_rect.dymax),
                base_layer=layer["comp"],
                metal_level=1,
            )
        )

        sub_imp = c.add_ref(
            gf.components.rectangle(
                size=(
                    sub_rect.dxsize + (2 * pp_enc_cmp),
                    cmp.dysize + (2 * pp_enc_cmp),
                ),
                layer=sub_imp_layer,
            )
        )
        sub_imp.dxmin = sub_rect.dxmin - pp_enc_cmp
        sub_imp.dymin = sub_rect.dymin - pp_enc_cmp

        # label generation
        if label == 1:
            c.add_label(
                sub_label,
                position=(
                    sub_con.dxmin + (sub_con.dxsize / 2),
                    sub_con.dymin + (sub_con.dysize / 2),
                ),
                layer=layer["metal1_label"],
            )

    return c


@gf.cell
def nplus_res(
    l_res: float = 0.1,
    w_res: float = 0.1,
    res_type: str = "nplus_s",
    sub: bool = False,
    deepnwell: bool = False,
    pcmpgr: bool = False,
    label: bool = False,
    r0_label: str = "",
    r1_label: str = "",
    sub_label: str = "",
) -> gf.Component:
    c = gf.Component()

    lvpwell_enc_cmp = 0.43
    dn_enc_lvpwell = 2.5
    sub_w = 0.36

    if res_type == "nplus_s":
        cmp_res_ext = 0.29
        con_enc = 0.07
    else:
        cmp_res_ext = 0.44
        con_enc = 0.0

    # adding res inst
    r_inst = c.add_ref(
        plus_res_inst(
            l_res=l_res,
            w_res=w_res,
            res_type=res_type,
            sub=sub,
            cmp_res_ext=cmp_res_ext,
            con_enc=con_enc,
            cmp_imp_layer=layer["nplus"],
            sub_imp_layer=layer["pplus"],
            label=label,
            r0_label=r0_label,
            r1_label=r1_label,
            sub_label=sub_label,
        )
    )

    if deepnwell == 1:
        lvpwell = c.add_ref(
            gf.components.rectangle(
                size=(
                    r_inst.dxsize + (2 * lvpwell_enc_cmp),
                    r_inst.dysize + (2 * lvpwell_enc_cmp),
                ),
                layer=layer["lvpwell"],
            )
        )
        lvpwell.dxmin = r_inst.dxmin - lvpwell_enc_cmp
        lvpwell.dymin = r_inst.dymin - lvpwell_enc_cmp

        dn_rect = c.add_ref(
            gf.components.rectangle(
                size=(
                    lvpwell.dxsize + (2 * dn_enc_lvpwell),
                    lvpwell.dysize + (2 * dn_enc_lvpwell),
                ),
                layer=layer["dnwell"],
            )
        )
        dn_rect.dxmin = lvpwell.dxmin - dn_enc_lvpwell
        dn_rect.dymin = lvpwell.dymin - dn_enc_lvpwell

        if pcmpgr == 1:
            c.add_ref(pcmpgr_gen(dn_rect=dn_rect, grw=sub_w))

    return c


@gf.cell
def pplus_res(
    l_res: float = 0.1,
    w_res: float = 0.1,
    res_type: str = "pplus_s",
    sub: bool = False,
    deepnwell: bool = False,
    pcmpgr: bool = False,
    label: bool = False,
    r0_label: str = "",
    r1_label: str = "",
    sub_label: str = "",
) -> gf.Component:
    c = gf.Component("res_dev")

    nw_enc_pcmp = 0.6
    dn_enc_ncmp = 0.66
    dn_enc_pcmp = 1.02
    sub_w = 0.36

    if res_type == "pplus_s":
        cmp_res_ext = 0.29
        con_enc = 0.07
    else:
        cmp_res_ext = 0.44
        con_enc = 0.0

    # adding res inst
    r_inst = c.add_ref(
        plus_res_inst(
            l_res=l_res,
            w_res=w_res,
            res_type=res_type,
            sub=1,
            cmp_res_ext=cmp_res_ext,
            con_enc=con_enc,
            cmp_imp_layer=layer["pplus"],
            sub_imp_layer=layer["nplus"],
            label=label,
            r0_label=r0_label,
            r1_label=r1_label,
            sub_label=sub_label,
        )
    )

    if deepnwell == 1:
        dn_rect = c.add_ref(
            gf.components.rectangle(
                size=(
                    r_inst.dxsize + (dn_enc_pcmp + dn_enc_ncmp),
                    r_inst.dysize + (2 * dn_enc_pcmp),
                ),
                layer=layer["dnwell"],
            )
        )
        dn_rect.dxmax = r_inst.dxmax + dn_enc_pcmp
        dn_rect.dymin = r_inst.dymin - dn_enc_pcmp

        if pcmpgr == 1:
            c.add_ref(pcmpgr_gen(dn_rect=dn_rect, grw=sub_w))

    else:
        nw_rect = c.add_ref(
            gf.components.rectangle(
                size=(
                    r_inst.dxsize + (2 * nw_enc_pcmp),
                    r_inst.dysize + (2 * nw_enc_pcmp),
                ),
                layer=layer["nwell"],
            )
        )
        nw_rect.dxmin = r_inst.dxmin - nw_enc_pcmp
        nw_rect.dymin = r_inst.dymin - nw_enc_pcmp

    return c


@gf.cell
def polyf_res_inst(
    l_res: float = 0.1,
    w_res: float = 0.1,
    res_type: str = "npolyf_s",
    pl_res_ext: float = 0.1,
    con_enc: float = 0.1,
    pl_imp_layer: LayerSpec = layer["nplus"],
    sub_imp_layer: LayerSpec = layer["pplus"],
    label: bool = False,
    r0_label: str = "",
    r1_label: str = "",
    sub_label: str = "",
) -> gf.Component:
    c = gf.Component()

    sub_w: float = 0.36
    np_enc_poly2 = 0.3
    pp_enc_cmp: float = 0.16
    comp_spacing: float = 0.72
    sab_res_ext = 0.28

    res_mk = c.add_ref(
        gf.components.rectangle(size=(l_res, w_res), layer=layer["res_mk"])
    )

    if "polyf_u" in res_type:
        sab_rect = c.add_ref(
            gf.components.rectangle(
                size=(res_mk.dxsize, res_mk.dysize + (2 * sab_res_ext)),
                layer=layer["sab"],
            )
        )
        sab_rect.dxmin = res_mk.dxmin
        sab_rect.dymin = res_mk.dymin - sab_res_ext

    pl = c.add_ref(
        gf.components.rectangle(
            size=(res_mk.dxsize + (2 * pl_res_ext), res_mk.dysize),
            layer=layer["poly2"],
        )
    )
    pl.dxmin = res_mk.dxmin - pl_res_ext
    pl.dymin = res_mk.dymin

    pl_con = via_stack(
        x_range=(pl.dxmin, res_mk.dxmin + con_enc),
        y_range=(pl.dymin, pl.dymax),
        base_layer=layer["poly2"],
        metal_level=1,
    )

    pl_con_arr = c.add_ref(
        component=pl_con,
        rows=1,
        columns=2,
        column_pitch=pl_res_ext - con_enc + res_mk.dxsize,
    )  # comp contact array

    pl_imp = c.add_ref(
        gf.components.rectangle(
            size=(pl.dxsize + (2 * np_enc_poly2), pl.dysize + (2 * np_enc_poly2)),
            layer=pl_imp_layer,
        )
    )
    pl_imp.dxmin = pl.dxmin - np_enc_poly2
    pl_imp.dymin = pl.dymin - np_enc_poly2

    sub_rect = c.add_ref(
        gf.components.rectangle(size=(sub_w, w_res), layer=layer["comp"])
    )
    sub_rect.dxmax = pl.dxmin - comp_spacing
    sub_rect.dymin = pl.dymin

    # sub_rect contact
    sub_con = c.add_ref(
        via_stack(
            x_range=(sub_rect.dxmin, sub_rect.dxmax),
            y_range=(sub_rect.dymin, sub_rect.dymax),
            base_layer=layer["comp"],
            metal_level=1,
        )
    )

    sub_imp = c.add_ref(
        gf.components.rectangle(
            size=(
                sub_rect.dxsize + (2 * pp_enc_cmp),
                pl.dysize + (2 * pp_enc_cmp),
            ),
            layer=sub_imp_layer,
        )
    )
    sub_imp.dxmin = sub_rect.dxmin - pp_enc_cmp
    sub_imp.dymin = sub_rect.dymin - pp_enc_cmp

    # labels generation
    if label == 1:
        c.add_label(
            r0_label,
            position=(
                pl_con_arr.dxmin + (pl_con.dxsize / 2),
                pl_con_arr.dymin + (pl_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )
        c.add_label(
            r1_label,
            position=(
                pl_con_arr.dxmax - (pl_con.dxsize / 2),
                pl_con_arr.dymin + (pl_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

        c.add_label(
            sub_label,
            position=(
                sub_con.dxmin + (sub_con.dxsize / 2),
                sub_con.dymin + (sub_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

    return c


@gf.cell
def npolyf_res(
    l_res: float = 0.1,
    w_res: float = 0.1,
    res_type: str = "npolyf_s",
    deepnwell: bool = False,
    pcmpgr: bool = False,
    label: bool = False,
    r0_label: str = "",
    r1_label: str = "",
    sub_label: str = "",
) -> gf.Component:
    c = gf.Component("res_dev")

    lvpwell_enc_cmp = 0.43
    dn_enc_lvpwell = 2.5
    sub_w = 0.36

    if res_type == "npolyf_s":
        pl_res_ext = 0.29
        con_enc = 0.07
    else:
        pl_res_ext = 0.44
        con_enc = 0.0

    # adding res inst
    r_inst = c.add_ref(
        polyf_res_inst(
            l_res=l_res,
            w_res=w_res,
            res_type=res_type,
            pl_res_ext=pl_res_ext,
            con_enc=con_enc,
            pl_imp_layer=layer["nplus"],
            sub_imp_layer=layer["pplus"],
            label=label,
            r0_label=r0_label,
            r1_label=r1_label,
            sub_label=sub_label,
        )
    )

    if deepnwell == 1:
        lvpwell = c.add_ref(
            gf.components.rectangle(
                size=(
                    r_inst.dxsize + (2 * lvpwell_enc_cmp),
                    r_inst.dysize + (2 * lvpwell_enc_cmp),
                ),
                layer=layer["lvpwell"],
            )
        )
        lvpwell.dxmin = r_inst.dxmin - lvpwell_enc_cmp
        lvpwell.dymin = r_inst.dymin - lvpwell_enc_cmp

        dn_rect = c.add_ref(
            gf.components.rectangle(
                size=(
                    lvpwell.dxsize + (2 * dn_enc_lvpwell),
                    lvpwell.dysize + (2 * dn_enc_lvpwell),
                ),
                layer=layer["dnwell"],
            )
        )
        dn_rect.dxmin = lvpwell.dxmin - dn_enc_lvpwell
        dn_rect.dymin = lvpwell.dymin - dn_enc_lvpwell

        if pcmpgr == 1:
            c.add_ref(pcmpgr_gen(dn_rect=dn_rect, grw=sub_w))

    return c


@gf.cell
def ppolyf_res(
    l_res: float = 0.1,
    w_res: float = 0.1,
    res_type: str = "ppolyf_s",
    deepnwell: bool = False,
    pcmpgr: bool = False,
    label: bool = False,
    r0_label: str = "",
    r1_label: str = "",
    sub_label: str = "",
) -> gf.Component:
    c = gf.Component("res_dev")

    sub_w = 0.36
    dn_enc_ncmp = 0.66
    dn_enc_poly2 = 1.34

    if res_type == "ppolyf_s":
        pl_res_ext = 0.29
        con_enc = 0.07
    else:
        pl_res_ext = 0.44
        con_enc = 0.0

    if deepnwell == 1:
        sub_layer = layer["nplus"]
    else:
        sub_layer = layer["pplus"]

    # adding res inst
    r_inst = c.add_ref(
        polyf_res_inst(
            l_res=l_res,
            w_res=w_res,
            res_type=res_type,
            pl_res_ext=pl_res_ext,
            con_enc=con_enc,
            pl_imp_layer=layer["pplus"],
            sub_imp_layer=sub_layer,
            label=label,
            r0_label=r0_label,
            r1_label=r1_label,
            sub_label=sub_label,
        )
    )

    if deepnwell == 1:
        dn_rect = c.add_ref(
            gf.components.rectangle(
                size=(
                    r_inst.dxsize + (dn_enc_poly2 + dn_enc_ncmp),
                    r_inst.dysize + (2 * dn_enc_poly2),
                ),
                layer=layer["dnwell"],
            )
        )
        dn_rect.dxmax = r_inst.dxmax + dn_enc_poly2
        dn_rect.dymin = r_inst.dymin - dn_enc_poly2

        if pcmpgr == 1:
            c.add_ref(pcmpgr_gen(dn_rect=dn_rect, grw=sub_w))

    return c


@gf.cell
def ppolyf_u_high_Rs_res(
    l_res: float = 0.42,
    w_res: float = 0.42,
    volt: str = "3.3V",
    deepnwell: bool = False,
    pcmpgr: bool = False,
    label: bool = False,
    r0_label: str = "",
    r1_label: str = "",
    sub_label: str = "",
) -> gf.Component:
    c = gf.Component("res_dev")

    dn_enc_ncmp = 0.62
    dn_enc_poly2 = 1.34

    pl_res_ext = 0.64

    sub_w: float = 0.42
    pp_enc_poly2 = 0.18
    pp_enc_cmp: float = 0.02
    comp_spacing: float = 0.7
    sab_res_ext = (0.1, 0.28)
    con_size = 0.36
    resis_enc = (1.04, 0.4)
    dg_enc_dn = 0.5

    res_mk = c.add_ref(
        gf.components.rectangle(size=(l_res, w_res), layer=layer["res_mk"])
    )

    resis_mk = c.add_ref(
        gf.components.rectangle(
            size=(
                res_mk.dxsize + (2 * resis_enc[0]),
                res_mk.dysize + (2 * resis_enc[1]),
            ),
            layer=layer["resistor"],
        )
    )

    resis_mk.dxmin = res_mk.dxmin - resis_enc[0]
    resis_mk.dymin = res_mk.dymin - resis_enc[1]

    sab_rect = c.add_ref(
        gf.components.rectangle(
            size=(
                res_mk.dxsize + (2 * sab_res_ext[0]),
                res_mk.dysize + (2 * sab_res_ext[1]),
            ),
            layer=layer["sab"],
        )
    )
    sab_rect.dxmin = res_mk.dxmin - sab_res_ext[0]
    sab_rect.dymin = res_mk.dymin - sab_res_ext[1]

    pl = c.add_ref(
        gf.components.rectangle(
            size=(res_mk.dxsize + (2 * pl_res_ext), res_mk.dysize),
            layer=layer["poly2"],
        )
    )
    pl.dxmin = res_mk.dxmin - pl_res_ext
    pl.dymin = res_mk.dymin

    pl_con = via_stack(
        x_range=(pl.dxmin, pl.dxmin + con_size),
        y_range=(pl.dymin, pl.dymax),
        base_layer=layer["poly2"],
        metal_level=1,
    )

    pl_con_arr = c.add_ref(
        component=pl_con,
        rows=1,
        columns=2,
        column_pitch=(pl.dxsize - con_size),
    )  # comp contact array

    pplus = gf.components.rectangle(
        size=(pl_res_ext + pp_enc_poly2, pl.dysize + (2 * pp_enc_poly2)),
        layer=layer["pplus"],
    )

    pplus_arr = c.add_ref(
        component=pplus, rows=1, columns=2, column_pitch=(pplus.dxsize + res_mk.dxsize)
    )

    pplus_arr.dxmin = pl.dxmin - pp_enc_poly2
    pplus_arr.dymin = pl.dymin - pp_enc_poly2

    sub_rect = c.add_ref(
        gf.components.rectangle(size=(sub_w, w_res), layer=layer["comp"])
    )
    sub_rect.dxmax = pl.dxmin - comp_spacing
    sub_rect.dymin = pl.dymin

    # sub_rect contact
    sub_con = c.add_ref(
        via_stack(
            x_range=(sub_rect.dxmin, sub_rect.dxmax),
            y_range=(sub_rect.dymin, sub_rect.dymax),
            base_layer=layer["comp"],
            metal_level=1,
        )
    )

    # labels generation
    if label == 1:
        c.add_label(
            r0_label,
            position=(
                pl_con_arr.dxmin + (pl_con.dxsize / 2),
                pl_con_arr.dymin + (pl_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )
        c.add_label(
            r1_label,
            position=(
                pl_con_arr.dxmax - (pl_con.dxsize / 2),
                pl_con_arr.dymin + (pl_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

        c.add_label(
            sub_label,
            position=(
                sub_con.dxmin + (sub_con.dxsize / 2),
                sub_con.dymin + (sub_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

    if deepnwell == 1:
        sub_layer = layer["nplus"]
    else:
        sub_layer = layer["pplus"]

    sub_imp = c.add_ref(
        gf.components.rectangle(
            size=(
                sub_rect.dxsize + (2 * pp_enc_cmp),
                pl.dysize + (2 * pp_enc_cmp),
            ),
            layer=sub_layer,
        )
    )
    sub_imp.dxmin = sub_rect.dxmin - pp_enc_cmp
    sub_imp.dymin = sub_rect.dymin - pp_enc_cmp

    if deepnwell == 1:
        dn_rect = c.add_ref(
            gf.components.rectangle(
                size=(
                    (pl.dxmax - sub_rect.dxmin) + (dn_enc_poly2 + dn_enc_ncmp),
                    pl.dysize + (2 * dn_enc_poly2),
                ),
                layer=layer["dnwell"],
            )
        )
        dn_rect.dxmax = pl.dxmax + dn_enc_poly2
        dn_rect.dymin = pl.dymin - dn_enc_poly2

        if volt == "5/6V":
            dg = c.add_ref(
                gf.components.rectangle(
                    size=(
                        dn_rect.dxsize + (2 * dg_enc_dn),
                        dn_rect.dysize + (2 * dg_enc_dn),
                    ),
                    layer=layer["dualgate"],
                )
            )

            dg.dxmin = dn_rect.dxmin - dg_enc_dn
            dg.dymin = dn_rect.dymin - dg_enc_dn

        if pcmpgr == 1:
            c.add_ref(pcmpgr_gen(dn_rect=dn_rect, grw=sub_w))

    else:
        if volt == "5/6V":
            dg = c.add_ref(
                gf.components.rectangle(
                    size=(resis_mk.dxsize, resis_mk.dysize), layer=layer["dualgate"]
                )
            )

            dg.dxmin = resis_mk.dxmin
            dg.dymin = resis_mk.dymin

    return c


@gf.cell
def well_res(
    l_res: float = 0.42,
    w_res: float = 0.42,
    res_type: str = "nwell",
    pcmpgr: bool = False,
    label: bool = False,
    r0_label: str = "",
    r1_label: str = "",
    sub_label: str = "",
) -> gf.Component:
    c = gf.Component("res_dev")

    nw_res_ext = 0.48
    nw_res_enc = 0.5
    nw_enc_cmp = 0.12

    sub_w: float = 0.36
    pp_enc_cmp: float = 0.16
    nw_comp_spacing: float = 0.72
    dn_enc_lvpwell = 2.5

    if res_type == "pwell":
        cmp_imp_layer = layer["pplus"]
        sub_imp_layer = layer["nplus"]
        well_layer = layer["lvpwell"]
    else:
        cmp_imp_layer = layer["nplus"]
        sub_imp_layer = layer["pplus"]
        well_layer = layer["nwell"]

    res_mk = c.add_ref(
        gf.components.rectangle(
            size=(l_res, w_res + (2 * nw_res_enc)), layer=layer["res_mk"]
        )
    )

    well_rect = c.add_ref(
        gf.components.rectangle(
            size=(res_mk.dxsize + (2 * nw_res_ext), w_res), layer=well_layer
        )
    )
    well_rect.dxmin = res_mk.dxmin - nw_res_ext
    well_rect.dymin = res_mk.dymin + nw_res_enc

    @gf.cell
    def comp_related_gen(size: Float2 = (0.42, 0.42)) -> gf.Component:
        c = gf.Component()

        cmp = c.add_ref(gf.components.rectangle(size=size, layer=layer["comp"]))
        cmp.dxmin = well_rect.dxmin + nw_enc_cmp
        cmp.dymin = well_rect.dymin + nw_enc_cmp

        c.add_ref(
            via_stack(
                x_range=(cmp.dxmin, cmp.dxmax),
                y_range=(cmp.dymin, cmp.dymax),
                base_layer=layer["comp"],
                metal_level=1,
            )
        )  # contact

        return c

    con_polys = comp_related_gen(
        size=(
            res_mk.dxmin - well_rect.dxmin - nw_enc_cmp,
            well_rect.dysize - (2 * nw_enc_cmp),
        )
    )

    con_polys_arr = c.add_ref(
        component=con_polys,
        rows=1,
        columns=2,
        column_pitch=(well_rect.dxsize - (2 * nw_enc_cmp) - con_polys.dxsize),
    )  # comp and its related contact array

    nplus_rect = gf.components.rectangle(
        size=(
            con_polys.dxsize + (2 * pp_enc_cmp),
            con_polys.dysize + (2 * pp_enc_cmp),
        ),
        layer=cmp_imp_layer,
    )
    nplus_arr = c.add_ref(
        component=nplus_rect,
        rows=1,
        columns=2,
        column_pitch=(well_rect.dxsize - (2 * nw_enc_cmp) - con_polys.dxsize),
    )
    nplus_arr.dxmin = con_polys.dxmin - pp_enc_cmp
    nplus_arr.dymin = con_polys.dymin - pp_enc_cmp

    sub_rect = c.add_ref(
        gf.components.rectangle(size=(sub_w, well_rect.dysize), layer=layer["comp"])
    )
    sub_rect.dxmax = well_rect.dxmin - nw_comp_spacing
    sub_rect.dymin = well_rect.dymin

    # sub_rect contact
    sub_con = c.add_ref(
        via_stack(
            x_range=(sub_rect.dxmin, sub_rect.dxmax),
            y_range=(sub_rect.dymin, sub_rect.dymax),
            base_layer=layer["comp"],
            metal_level=1,
        )
    )

    sub_imp = c.add_ref(
        gf.components.rectangle(
            size=(
                sub_rect.dxsize + (2 * pp_enc_cmp),
                well_rect.dysize + (2 * pp_enc_cmp),
            ),
            layer=sub_imp_layer,
        )
    )
    sub_imp.dxmin = sub_rect.dxmin - pp_enc_cmp
    sub_imp.dymin = sub_rect.dymin - pp_enc_cmp

    if res_type == "pwell":
        dn_rect = c.add_ref(
            gf.components.rectangle(
                size=(
                    well_rect.dxsize + (2 * dn_enc_lvpwell),
                    well_rect.dysize + (2 * dn_enc_lvpwell),
                ),
                layer=layer["dnwell"],
            )
        )
        dn_rect.dxmin = well_rect.dxmin - dn_enc_lvpwell
        dn_rect.dymin = well_rect.dymin - dn_enc_lvpwell

        if pcmpgr == 1:
            c.add_ref(pcmpgr_gen(dn_rect=dn_rect, grw=sub_w))

    # labels generation
    if label == 1:
        c.add_label(
            r0_label,
            position=(
                con_polys_arr.dxmin + (con_polys.dxsize / 2),
                con_polys_arr.dymin + (con_polys.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )
        c.add_label(
            r1_label,
            position=(
                con_polys_arr.dxmax - (con_polys.dxsize / 2),
                con_polys_arr.dymin + (con_polys.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

        c.add_label(
            sub_label,
            position=(
                sub_con.dxmin + (sub_con.dxsize / 2),
                sub_con.dymin + (sub_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )
    return c


if __name__ == "__main__":
    c = res()
    c.show()



from math import ceil, floor
from typing import Any

import gdsfactory as gf
from gdsfactory.typings import Float2, LayerSpec, Size, Spacing

from gf180mcu.layers import layer


@gf.cell
def via_generator(
    x_range: Float2 = (0, 1),
    y_range: Float2 = (0, 1),
    via_size: Float2 = (0.17, 0.17),
    via_layer: LayerSpec = (66, 44),
    via_enclosure: Float2 = (0.06, 0.06),
    via_spacing: Float2 = (0.17, 0.17),
) -> gf.Component:
    """Return only vias withen the range xrange and yrange while enclosing by via_enclosure
    and set number of rows and number of columns according to ranges and via size and spacing.

    Args:
        x_range: dx range.
        y_range: dy range.
        via_size: via size.
        via_layer: via layer.
        via_enclosure: via enclosure.
        via_spacing: via spacing.

    """
    c = gf.Component()

    width = x_range[1] - x_range[0]
    length = y_range[1] - y_range[0]
    nr = floor(length / (via_size[1] + via_spacing[1]))
    if (length - nr * via_size[1] - (nr - 1) * via_spacing[1]) / 2 < via_enclosure[1]:
        nr -= 1

    nr = max(nr, 1)
    nc = ceil(width / (via_size[0] + via_spacing[0]))

    if (
        round(width - nc * via_size[0] - (nc - 1) * via_spacing[0], 2)
    ) / 2 < via_enclosure[0]:
        nc -= 1

    nc = max(nc, 1)
    via_sp = (via_size[0] + via_spacing[0], via_size[1] + via_spacing[1])

    rect_via = gf.components.rectangle(size=via_size, layer=via_layer)

    via_arr = c.add_ref(
        rect_via, rows=nr, columns=nc, column_pitch=via_sp[0], row_pitch=via_sp[1]
    )

    via_arr.dmove((x_range[0], y_range[0]))

    via_arr.dmovex((width - nc * via_size[0] - (nc - 1) * via_spacing[0]) / 2)
    via_arr.dmovey((length - nr * via_size[1] - (nr - 1) * via_spacing[1]) / 2)
    return c


@gf.cell
def via_stack(
    x_range: Float2 = (0, 1),
    y_range: Float2 = (0, 1),
    metal_level: int = 1,
    con_size: Size = (0.22, 0.22),
    con_enc: float = 0.07,
    m_enc: float = 0.06,
    con_spacing: Spacing = (0.28, 0.28),
    via_size: Size = (0.22, 0.22),
    via_spacing: Spacing = (0.28, 0.28),
    via_enc: Float2 = (0.06, 0.06),
    base_layer: LayerSpec = layer["metal1"],
    **kwargs: Any,
) -> gf.Component:
    """Returns a via stack withen the range xrange and yrange and expecting the base_layer to be drawen.

    Args:
        x_range: dx range.
        y_range: dy range.
        metal_level: metal level.
        con_size: contact size.
        con_enc: contact enclosure.
        m_enc: metal enclosure.
        con_spacing: contact spacing.
        via_size: via size.
        via_spacing: via spacing.
        via_enc: via enclosure.


    return via stack till the metal level indicated where :
    metal_level 1 : till m1
    metal_level 2 : till m2
    metal_level 3 : till m3
    metal_level 4 : till m4
    metal_level 5 : till m5
    withen the range xrange and yrange and expecting the base_layer to be drawen

    """
    c = gf.Component()
    if metal_level >= 1:
        con_gen = via_generator(
            x_range=x_range,
            y_range=y_range,
            via_size=con_size,
            via_enclosure=(con_enc, con_enc),
            via_layer=layer["contact"],
            via_spacing=con_spacing,
        )
        con = c.add_ref(con_gen)
        m1_x = con.dxsize + 2 * m_enc
        m1_y = con.dysize + 2 * m_enc
        m1 = c.add_ref(gf.components.rectangle(size=(m1_x, m1_y), layer=base_layer))
        m1.dxmin = con.dxmin - m_enc
        m1.dymin = con.dymin - m_enc

    if metal_level >= 2:
        via1_gen = via_generator(
            x_range=(m1.dxmin, m1.dxmax),
            y_range=(m1.dymin, m1.dymax),
            via_size=via_size,
            via_enclosure=via_enc,
            via_layer=layer["via1"],
            via_spacing=via_spacing,
        )
        via1 = c.add_ref(via1_gen)

        if (via1.dxmax - via1.dxmin + 2 * m_enc[0]) < (
            via_size[0] + 2 * via_enc[0]
        ) and metal_level >= 3:
            m2_x = via_size[0] + 2 * via_enc[0]

        else:
            m2_x = via1.dxmax - via1.dxmin + 2 * m_enc[0]

        if (via1.dymax - via1.dymin + 2 * m_enc[1]) < (
            via_size[1] + 2 * via_enc[1]
        ) and metal_level >= 3:
            m2_y = via_size[1] + 2 * via_enc[1]

        else:
            m2_y = via1.dymax - via1.dymin + 2 * m_enc[1]

        m2_mx = (m2_x - (via1.dxmax - via1.dxmin)) / 2
        m2_my = (m2_y - (via1.dymax - via1.dymin)) / 2

        m2 = c.add_ref(
            gf.components.rectangle(size=(m2_x, m2_y), layer=layer["metal2"])
        )
        m2.dmove((via1.dxmin - m2_mx, via1.dymin - m2_my))

    return c


if __name__ == "__main__":
    c = via_stack()
    c.show()
    # c = vias_gen_draw(start_layer="li",end_layer="poly")
    # c.show()



import gdsfactory as gf

from gf180mcu.layers import layer
from gf180mcu.via_generator import via_generator


@gf.cell
def cap_mim(
    mim_option: str = "A",
    metal_level: str = "M4",
    lc: float = 2,
    wc: float = 2,
    label: bool = False,
    top_label: str = "",
    bot_label: str = "",
) -> gf.Component:
    """Return mim cap.

    Args:
        min_option: MIM-A or MIM-B.
        metal_level: metal level. M4, M5, M6.
        lc: cap length.
        wc: cap width.
        label: 1 to add labels.
        top_label: top label.
        bot_label: bottom label.

    """
    c = gf.Component("mim_cap_dev")
    # used dimensions and layers

    # MIM Option selection
    if mim_option == "MIM-A":
        upper_layer = layer["metal3"]
        bottom_layer = layer["metal2"]
        via_layer = layer["via2"]
        up_label_layer = layer["metal3_label"]
        bot_label_layer = layer["metal2_label"]

    elif mim_option == "MIM-B":
        if metal_level == "M4":
            upper_layer = layer["metal4"]
            bottom_layer = layer["metal3"]
            via_layer = layer["via3"]
            up_label_layer = layer["metal4_label"]
            bot_label_layer = layer["metal3_label"]
        elif metal_level == "M5":
            upper_layer = layer["metal5"]
            bottom_layer = layer["metal4"]
            via_layer = layer["via4"]
            up_label_layer = layer["metal5_label"]
            bot_label_layer = layer["metal4_label"]
        elif metal_level == "M6":
            upper_layer = layer["metaltop"]
            bottom_layer = layer["metal5"]
            via_layer = layer["via5"]
            up_label_layer = layer["metaltop_label"]
            bot_label_layer = layer["metal5_label"]
    else:
        upper_layer = layer["metal3"]
        bottom_layer = layer["metal2"]
        via_layer = layer["via2"]
        up_label_layer = layer["metal3_label"]
        bot_label_layer = layer["metal2_label"]

    via_size = (0.22, 0.22)
    via_spacing = (0.5, 0.5)
    via_enc = (0.4, 0.4)

    bot_enc_top = 0.6
    l_mk_w = 0.1

    # drawing cap identifier and bottom , upper layers

    m_up = c.add_ref(
        gf.components.rectangle(
            size=(wc, lc),
            layer=upper_layer,
        )
    )

    fusetop = c.add_ref(
        gf.components.rectangle(size=(m_up.dxsize, m_up.dysize), layer=layer["fusetop"])
    )
    fusetop.dxmin = m_up.dxmin
    fusetop.dymin = m_up.dymin

    mim_l_mk = c.add_ref(
        gf.components.rectangle(size=(fusetop.dxsize, l_mk_w), layer=layer["mim_l_mk"])
    )
    mim_l_mk.dxmin = fusetop.dxmin
    mim_l_mk.dymin = fusetop.dymin

    m_dn = c.add_ref(
        gf.components.rectangle(
            size=(m_up.dxsize + (2 * bot_enc_top), m_up.dysize + (2 * bot_enc_top)),
            layer=bottom_layer,
        )
    )
    m_dn.dxmin = m_up.dxmin - bot_enc_top
    m_dn.dymin = m_up.dymin - bot_enc_top

    cap_mk = c.add_ref(
        gf.components.rectangle(size=(m_dn.dxsize, m_dn.dysize), layer=layer["cap_mk"])
    )
    cap_mk.dxmin = m_dn.dxmin
    cap_mk.dymin = m_dn.dymin

    # generatingg labels
    if label == 1:
        c.add_label(
            top_label,
            position=(m_up.dxmin + (m_up.dxsize / 2), m_dn.dxmin + (m_dn.dysize / 2)),
            layer=up_label_layer,
        )

        c.add_label(
            bot_label,
            position=(
                m_dn.dxmin + (m_dn.dxsize / 2),
                m_dn.dymin + (m_up.dymin - m_dn.dymin) / 2,
            ),
            layer=bot_label_layer,
        )

    # generatingg vias
    via = via_generator(
        x_range=(m_up.dxmin, m_up.dxmax),
        y_range=(m_up.dymin, m_up.dymax),
        via_enclosure=via_enc,
        via_layer=via_layer,
        via_size=via_size,
        via_spacing=via_spacing,
    )
    c.add_ref(via)
    return c


if __name__ == "__main__":
    c = cap_mim()
    c.show()



import gdsfactory as gf
from gdsfactory.typings import Float2

from gf180mcu.layers import layer
from gf180mcu.via_generator import via_generator, via_stack


@gf.cell
def diode_nd2ps(
    la: float = 0.1,
    wa: float = 0.1,
    cw: float = 0.1,
    volt: str = "3.3V",
    deepnwell: bool = False,
    pcmpgr: bool = False,
    label: bool = False,
    p_label: str = "",
    n_label: str = "",
) -> gf.Component:
    """Draw N+/LVPWELL diode (Outside DNWELL) by specifying parameters.

    Args::
        la: Float of diff length (anode).
        wa: Float of diff width (anode).
        cw: Float of cathode width.
        volt: String of operating voltage of the diode [3.3V, 5V/6V].
        deepnwell: Boolean of using Deep NWELL device.
        pcmpgr : Boolean of using P+ Guard Ring for Deep NWELL devices only.
        label: Boolean of adding labels.
    """
    c = gf.Component("diode_nd2ps_dev")

    comp_spacing: float = 0.48
    np_enc_comp: float = 0.16
    pp_enc_comp: float = 0.16

    con_size = 0.22
    con_sp = 0.28
    con_comp_enc = 0.07

    dg_enc_cmp = 0.24
    dn_enc_lvpwell = 2.5
    lvpwell_enc_ncmp = 0.6
    lvpwell_enc_pcmp = 0.16
    pcmpgr_enc_dn = 2.5

    # n generation
    ncmp = c.add_ref(gf.components.rectangle(size=(wa, la), layer=layer["comp"]))
    nplus = c.add_ref(
        gf.components.rectangle(
            size=(
                ncmp.dxsize + (2 * np_enc_comp),
                ncmp.dysize + (2 * np_enc_comp),
            ),
            layer=layer["nplus"],
        )
    )
    nplus.dxmin = ncmp.dxmin - np_enc_comp
    nplus.dymin = ncmp.dymin - np_enc_comp
    diode_mk = c.add_ref(
        gf.components.rectangle(
            size=(ncmp.dxsize, ncmp.dysize), layer=layer["diode_mk"]
        )
    )
    diode_mk.dxmin = ncmp.dxmin
    diode_mk.dymin = ncmp.dymin

    ncmp_con = c.add_ref(
        via_stack(
            x_range=(ncmp.dxmin, ncmp.dxmax),
            y_range=(ncmp.dymin, ncmp.dymax),
            base_layer=layer["comp"],
            metal_level=1,
        )
    )  # ncomp_con

    # p generation
    pcmp = c.add_ref(gf.components.rectangle(size=(cw, la), layer=layer["comp"]))
    pcmp.dxmax = ncmp.dxmin - comp_spacing
    pplus = c.add_ref(
        gf.components.rectangle(
            size=(
                pcmp.dxsize + (2 * pp_enc_comp),
                pcmp.dysize + (2 * pp_enc_comp),
            ),
            layer=layer["pplus"],
        )
    )
    pplus.dxmin = pcmp.dxmin - pp_enc_comp
    pplus.dymin = pcmp.dymin - pp_enc_comp

    pcmp_con = c.add_ref(
        via_stack(
            x_range=(pcmp.dxmin, pcmp.dxmax),
            y_range=(pcmp.dymin, pcmp.dymax),
            base_layer=layer["comp"],
            metal_level=1,
        )
    )  # pcomp_con

    # labels generation
    if label == 1:
        # n_label generation
        c.add_label(
            n_label,
            position=(
                ncmp_con.dxmin + (ncmp_con.dxsize / 2),
                ncmp_con.dymin + (ncmp_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

        # p_label generation
        c.add_label(
            p_label,
            position=(
                pcmp_con.dxmin + (pcmp_con.dxsize / 2),
                pcmp_con.dymin + (pcmp_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

    if volt == "5/6V":
        dg = c.add_ref(
            gf.components.rectangle(
                size=(
                    ncmp.dxmax - pcmp.dxmin + (2 * dg_enc_cmp),
                    ncmp.dysize + (2 * dg_enc_cmp),
                ),
                layer=layer["dualgate"],
            )
        )
        dg.dxmin = pcmp.dxmin - dg_enc_cmp
        dg.dymin = pcmp.dymin - dg_enc_cmp

    if deepnwell == 1:
        lvpwell = c.add_ref(
            gf.components.rectangle(
                size=(
                    ncmp.dxmax - pcmp.dxmin + (lvpwell_enc_ncmp + lvpwell_enc_pcmp),
                    ncmp.dysize + (2 * lvpwell_enc_ncmp),
                ),
                layer=layer["lvpwell"],
            )
        )

        lvpwell.dxmin = pcmp.dxmin - lvpwell_enc_pcmp
        lvpwell.dymin = ncmp.dymin - lvpwell_enc_ncmp

        dn_rect = c.add_ref(
            gf.components.rectangle(
                size=(
                    lvpwell.dxsize + (2 * dn_enc_lvpwell),
                    lvpwell.dysize + (2 * dn_enc_lvpwell),
                ),
                layer=layer["dnwell"],
            )
        )

        dn_rect.dxmin = lvpwell.dxmin - dn_enc_lvpwell
        dn_rect.dymin = lvpwell.dymin - dn_enc_lvpwell

        if pcmpgr == 1:
            c_temp_gr = gf.Component("temp_store guard ring")
            rect_pcmpgr_in = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(
                        (dn_rect.dxmax - dn_rect.dxmin) + 2 * pcmpgr_enc_dn,
                        (dn_rect.dymax - dn_rect.dymin) + 2 * pcmpgr_enc_dn,
                    ),
                    layer=layer["comp"],
                )
            )
            rect_pcmpgr_in.dmove(
                (dn_rect.dxmin - pcmpgr_enc_dn, dn_rect.dymin - pcmpgr_enc_dn)
            )
            rect_pcmpgr_out = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(
                        (rect_pcmpgr_in.dxmax - rect_pcmpgr_in.dxmin) + 2 * cw,
                        (rect_pcmpgr_in.dymax - rect_pcmpgr_in.dymin) + 2 * cw,
                    ),
                    layer=layer["comp"],
                )
            )
            rect_pcmpgr_out.dmove(
                (rect_pcmpgr_in.dxmin - cw, rect_pcmpgr_in.dymin - cw)
            )
            c.add_ref(
                gf.boolean(
                    A=rect_pcmpgr_out,
                    B=rect_pcmpgr_in,
                    operation="A-B",
                    layer=layer["comp"],
                )
            )  # guardring Bulk draw

            psdm_in = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(
                        (rect_pcmpgr_in.dxmax - rect_pcmpgr_in.dxmin) - 2 * pp_enc_comp,
                        (rect_pcmpgr_in.dymax - rect_pcmpgr_in.dymin) - 2 * pp_enc_comp,
                    ),
                    layer=layer["pplus"],
                )
            )
            psdm_in.dmove(
                (
                    rect_pcmpgr_in.dxmin + pp_enc_comp,
                    rect_pcmpgr_in.dymin + pp_enc_comp,
                )
            )
            psdm_out = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(
                        (rect_pcmpgr_out.dxmax - rect_pcmpgr_out.dxmin)
                        + 2 * pp_enc_comp,
                        (rect_pcmpgr_out.dymax - rect_pcmpgr_out.dymin)
                        + 2 * pp_enc_comp,
                    ),
                    layer=layer["pplus"],
                )
            )
            psdm_out.dmove(
                (
                    rect_pcmpgr_out.dxmin - pp_enc_comp,
                    rect_pcmpgr_out.dymin - pp_enc_comp,
                )
            )
            c.add_ref(
                gf.boolean(A=psdm_out, B=psdm_in, operation="A-B", layer=layer["pplus"])
            )  # psdm draw

            # generatingg contacts

            c.add_ref(
                via_generator(
                    x_range=(
                        rect_pcmpgr_in.dxmin + con_size,
                        rect_pcmpgr_in.dxmax - con_size,
                    ),
                    y_range=(rect_pcmpgr_out.dymin, rect_pcmpgr_in.dymin),
                    via_enclosure=(con_comp_enc, con_comp_enc),
                    via_layer=layer["contact"],
                    via_size=(con_size, con_size),
                    via_spacing=(con_sp, con_sp),
                )
            )  # bottom contact

            c.add_ref(
                via_generator(
                    x_range=(
                        rect_pcmpgr_in.dxmin + con_size,
                        rect_pcmpgr_in.dxmax - con_size,
                    ),
                    y_range=(rect_pcmpgr_in.dymax, rect_pcmpgr_out.dymax),
                    via_enclosure=(con_comp_enc, con_comp_enc),
                    via_layer=layer["contact"],
                    via_size=(con_size, con_size),
                    via_spacing=(con_sp, con_sp),
                )
            )  # upper contact

            c.add_ref(
                via_generator(
                    x_range=(rect_pcmpgr_out.dxmin, rect_pcmpgr_in.dxmin),
                    y_range=(
                        rect_pcmpgr_in.dymin + con_size,
                        rect_pcmpgr_in.dymax - con_size,
                    ),
                    via_enclosure=(con_comp_enc, con_comp_enc),
                    via_layer=layer["contact"],
                    via_size=(con_size, con_size),
                    via_spacing=(con_sp, con_sp),
                )
            )  # right contact

            c.add_ref(
                via_generator(
                    x_range=(rect_pcmpgr_in.dxmax, rect_pcmpgr_out.dxmax),
                    y_range=(
                        rect_pcmpgr_in.dymin + con_size,
                        rect_pcmpgr_in.dymax - con_size,
                    ),
                    via_enclosure=(con_comp_enc, con_comp_enc),
                    via_layer=layer["contact"],
                    via_size=(con_size, con_size),
                    via_spacing=(con_sp, con_sp),
                )
            )  # left contact

            comp_m1_in = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(rect_pcmpgr_in.dxsize, rect_pcmpgr_in.dysize),
                    layer=layer["metal1"],
                )
            )

            comp_m1_out = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(
                        (comp_m1_in.dxsize) + 2 * cw,
                        (comp_m1_in.dysize) + 2 * cw,
                    ),
                    layer=layer["metal1"],
                )
            )
            comp_m1_out.dmove((rect_pcmpgr_in.dxmin - cw, rect_pcmpgr_in.dymin - cw))
            c.add_ref(
                gf.boolean(
                    A=rect_pcmpgr_out,
                    B=rect_pcmpgr_in,
                    operation="A-B",
                    layer=layer["metal1"],
                )
            )  # guardring metal1

    return c


@gf.cell
def diode_pd2nw(
    la: float = 0.1,
    wa: float = 0.1,
    cw: float = 0.1,
    volt: str = "3.3V",
    deepnwell: bool = False,
    pcmpgr: bool = False,
    label: bool = False,
    p_label: str = "",
    n_label: str = "",
) -> gf.Component:
    """Usage:-
     used to draw 3.3V P+/Nwell diode (Outside DNWELL) by specifying parameters
    Arguments:-
     la         : Float of diffusion length (anode)
     wa         : Float of diffusion width (anode)
     volt       : String of operating voltage of the diode [3.3V, 5V/6V]
     deepnwell  : Boolean of using Deep NWELL device
     pcmpgr     : Boolean of using P+ Guard Ring for Deep NWELL devices only.
    """
    c = gf.Component("diode_pd2nw_dev")

    comp_spacing: float = 0.48
    np_enc_comp: float = 0.16
    pp_enc_comp: float = 0.16

    con_size = 0.22
    con_sp = 0.28
    con_comp_enc = 0.07

    dg_enc_cmp = 0.24
    dn_enc_nwell = 0.5
    nwell_ncmp_enc = 0.12
    nwell_pcmp_enc = 0.43
    pcmpgr_enc_dn = 2.5

    # p generation
    pcmp = c.add_ref(gf.components.rectangle(size=(wa, la), layer=layer["comp"]))
    pplus = c.add_ref(
        gf.components.rectangle(
            size=(
                pcmp.dxsize + (2 * pp_enc_comp),
                pcmp.dysize + (2 * pp_enc_comp),
            ),
            layer=layer["pplus"],
        )
    )
    pplus.dxmin = pcmp.dxmin - pp_enc_comp
    pplus.dymin = pcmp.dymin - pp_enc_comp
    diode_mk = c.add_ref(
        gf.components.rectangle(
            size=(pcmp.dxsize, pcmp.dysize), layer=layer["diode_mk"]
        )
    )
    diode_mk.dxmin = pcmp.dxmin
    diode_mk.dymin = pcmp.dymin

    pcmp_con = c.add_ref(
        via_stack(
            x_range=(pcmp.dxmin, pcmp.dxmax),
            y_range=(pcmp.dymin, pcmp.dymax),
            base_layer=layer["comp"],
            metal_level=1,
        )
    )  # pcomp_contact

    # n generation
    ncmp = c.add_ref(gf.components.rectangle(size=(cw, la), layer=layer["comp"]))
    ncmp.dxmax = pcmp.dxmin - comp_spacing
    nplus = c.add_ref(
        gf.components.rectangle(
            size=(
                ncmp.dxsize + (2 * np_enc_comp),
                ncmp.dysize + (2 * np_enc_comp),
            ),
            layer=layer["nplus"],
        )
    )
    nplus.dxmin = ncmp.dxmin - np_enc_comp
    nplus.dymin = ncmp.dymin - np_enc_comp

    ncmp_con = c.add_ref(
        via_stack(
            x_range=(ncmp.dxmin, ncmp.dxmax),
            y_range=(ncmp.dymin, ncmp.dymax),
            base_layer=layer["comp"],
            metal_level=1,
        )
    )  # ncomp contact

    # labels generation
    if label == 1:
        # n_label generation
        c.add_label(
            n_label,
            position=(
                ncmp_con.dxmin + (ncmp_con.dxsize / 2),
                ncmp_con.dymin + (ncmp_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

        # p_label generation
        c.add_label(
            p_label,
            position=(
                pcmp_con.dxmin + (pcmp_con.dxsize / 2),
                pcmp_con.dymin + (pcmp_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

    if volt == "5/6V":
        dg = c.add_ref(
            gf.components.rectangle(
                size=(
                    pcmp.dxmax - ncmp.dxmin + (2 * dg_enc_cmp),
                    ncmp.dysize + (2 * dg_enc_cmp),
                ),
                layer=layer["dualgate"],
            )
        )
        dg.dxmin = ncmp.dxmin - dg_enc_cmp
        dg.dymin = ncmp.dymin - dg_enc_cmp

    # nwell generation
    nwell = c.add_ref(
        gf.components.rectangle(
            size=(
                pcmp.dxmax - ncmp.dxmin + (nwell_ncmp_enc + nwell_pcmp_enc),
                pcmp.dysize + (2 * nwell_pcmp_enc),
            ),
            layer=layer["nwell"],
        )
    )

    nwell.dxmin = ncmp.dxmin - nwell_ncmp_enc
    nwell.dymin = pcmp.dymin - nwell_pcmp_enc

    if deepnwell == 1:
        dn_rect = c.add_ref(
            gf.components.rectangle(
                size=(
                    nwell.dxsize + (2 * dn_enc_nwell),
                    nwell.dysize + (2 * dn_enc_nwell),
                ),
                layer=layer["dnwell"],
            )
        )

        dn_rect.dxmin = nwell.dxmin - dn_enc_nwell
        dn_rect.dymin = nwell.dymin - dn_enc_nwell

        if pcmpgr == 1:
            c_temp_gr = gf.Component("temp_store guard ring")
            rect_pcmpgr_in = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(
                        (dn_rect.dxmax - dn_rect.dxmin) + 2 * pcmpgr_enc_dn,
                        (dn_rect.dymax - dn_rect.dymin) + 2 * pcmpgr_enc_dn,
                    ),
                    layer=layer["comp"],
                )
            )
            rect_pcmpgr_in.dmove(
                (dn_rect.dxmin - pcmpgr_enc_dn, dn_rect.dymin - pcmpgr_enc_dn)
            )
            rect_pcmpgr_out = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(
                        (rect_pcmpgr_in.dxmax - rect_pcmpgr_in.dxmin) + 2 * cw,
                        (rect_pcmpgr_in.dymax - rect_pcmpgr_in.dymin) + 2 * cw,
                    ),
                    layer=layer["comp"],
                )
            )
            rect_pcmpgr_out.dmove(
                (rect_pcmpgr_in.dxmin - cw, rect_pcmpgr_in.dymin - cw)
            )
            c.add_ref(
                gf.boolean(
                    A=rect_pcmpgr_out,
                    B=rect_pcmpgr_in,
                    operation="A-B",
                    layer=layer["comp"],
                )
            )  # Bulk guardring

            psdm_in = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(
                        (rect_pcmpgr_in.dxmax - rect_pcmpgr_in.dxmin) - 2 * pp_enc_comp,
                        (rect_pcmpgr_in.dymax - rect_pcmpgr_in.dymin) - 2 * pp_enc_comp,
                    ),
                    layer=layer["pplus"],
                )
            )
            psdm_in.dmove(
                (
                    rect_pcmpgr_in.dxmin + pp_enc_comp,
                    rect_pcmpgr_in.dymin + pp_enc_comp,
                )
            )
            psdm_out = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(
                        (rect_pcmpgr_out.dxmax - rect_pcmpgr_out.dxmin)
                        + 2 * pp_enc_comp,
                        (rect_pcmpgr_out.dymax - rect_pcmpgr_out.dymin)
                        + 2 * pp_enc_comp,
                    ),
                    layer=layer["pplus"],
                )
            )
            psdm_out.dmove(
                (
                    rect_pcmpgr_out.dxmin - pp_enc_comp,
                    rect_pcmpgr_out.dymin - pp_enc_comp,
                )
            )
            c.add_ref(
                gf.boolean(A=psdm_out, B=psdm_in, operation="A-B", layer=layer["pplus"])
            )  # psdm guardring

            # generatingg contacts

            c.add_ref(
                via_generator(
                    x_range=(
                        rect_pcmpgr_in.dxmin + con_size,
                        rect_pcmpgr_in.dxmax - con_size,
                    ),
                    y_range=(rect_pcmpgr_out.dymin, rect_pcmpgr_in.dymin),
                    via_enclosure=(con_comp_enc, con_comp_enc),
                    via_layer=layer["contact"],
                    via_size=(con_size, con_size),
                    via_spacing=(con_sp, con_sp),
                )
            )  # bottom contact

            c.add_ref(
                via_generator(
                    x_range=(
                        rect_pcmpgr_in.dxmin + con_size,
                        rect_pcmpgr_in.dxmax - con_size,
                    ),
                    y_range=(rect_pcmpgr_in.dymax, rect_pcmpgr_out.dymax),
                    via_enclosure=(con_comp_enc, con_comp_enc),
                    via_layer=layer["contact"],
                    via_size=(con_size, con_size),
                    via_spacing=(con_sp, con_sp),
                )
            )  # upper contact

            c.add_ref(
                via_generator(
                    x_range=(rect_pcmpgr_out.dxmin, rect_pcmpgr_in.dxmin),
                    y_range=(
                        rect_pcmpgr_in.dymin + con_size,
                        rect_pcmpgr_in.dymax - con_size,
                    ),
                    via_enclosure=(con_comp_enc, con_comp_enc),
                    via_layer=layer["contact"],
                    via_size=(con_size, con_size),
                    via_spacing=(con_sp, con_sp),
                )
            )  # right contact

            c.add_ref(
                via_generator(
                    x_range=(rect_pcmpgr_in.dxmax, rect_pcmpgr_out.dxmax),
                    y_range=(
                        rect_pcmpgr_in.dymin + con_size,
                        rect_pcmpgr_in.dymax - con_size,
                    ),
                    via_enclosure=(con_comp_enc, con_comp_enc),
                    via_layer=layer["contact"],
                    via_size=(con_size, con_size),
                    via_spacing=(con_sp, con_sp),
                )
            )  # left contact

            comp_m1_in = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(rect_pcmpgr_in.dxsize, rect_pcmpgr_in.dysize),
                    layer=layer["metal1"],
                )
            )

            comp_m1_out = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(
                        (comp_m1_in.dxsize) + 2 * cw,
                        (comp_m1_in.dysize) + 2 * cw,
                    ),
                    layer=layer["metal1"],
                )
            )
            comp_m1_out.dmove((rect_pcmpgr_in.dxmin - cw, rect_pcmpgr_in.dymin - cw))
            c.add_ref(
                gf.boolean(
                    A=rect_pcmpgr_out,
                    B=rect_pcmpgr_in,
                    operation="A-B",
                    layer=layer["metal1"],
                )
            )  # guardring metal1

    return c


@gf.cell
def diode_nw2ps(
    la: float = 0.1,
    wa: float = 0.1,
    cw: float = 0.1,
    volt: str = "3.3V",
    label: bool = False,
    p_label: str = "",
    n_label: str = "",
) -> gf.Component:
    """Used to draw 3.3V Nwell/Psub diode by specifying parameters.

    Args:
        la: anode length.
        wa: anode width.
        cw: cathode width.
        volt: operating voltage of the diode [3.3V, 5V/6V]

    """
    c = gf.Component()

    comp_spacing: float = 0.48
    np_enc_comp: float = 0.16
    pp_enc_comp: float = 0.16

    dg_enc_cmp = 0.24

    nwell_ncmp_enc = 0.16

    # n generation
    ncmp = c.add_ref(gf.components.rectangle(size=(wa, la), layer=layer["comp"]))
    nplus = c.add_ref(
        gf.components.rectangle(
            size=(
                ncmp.dxsize + (2 * np_enc_comp),
                ncmp.dysize + (2 * np_enc_comp),
            ),
            layer=layer["nplus"],
        )
    )
    nplus.dxmin = ncmp.dxmin - np_enc_comp
    nplus.dymin = ncmp.dymin - np_enc_comp
    diode_mk = c.add_ref(
        gf.components.rectangle(
            size=(ncmp.dxsize, ncmp.dysize), layer=layer["diode_mk"]
        )
    )
    diode_mk.dxmin = ncmp.dxmin
    diode_mk.dymin = ncmp.dymin

    nwell = c.add_ref(
        gf.components.rectangle(
            size=(
                ncmp.dxsize + (2 * nwell_ncmp_enc),
                ncmp.dysize + (2 * nwell_ncmp_enc),
            ),
            layer=layer["nwell"],
        )
    )
    nwell.dxmin = ncmp.dxmin - nwell_ncmp_enc
    nwell.dymin = ncmp.dymin - nwell_ncmp_enc

    n_con = c.add_ref(
        via_stack(
            x_range=(ncmp.dxmin, ncmp.dxmax),
            y_range=(ncmp.dymin, ncmp.dymax),
            base_layer=layer["comp"],
            metal_level=1,
        )
    )  # ncomp contact

    # p generation
    pcmp = c.add_ref(gf.components.rectangle(size=(cw, la), layer=layer["comp"]))
    pcmp.dxmax = ncmp.dxmin - comp_spacing
    pplus = c.add_ref(
        gf.components.rectangle(
            size=(
                pcmp.dxsize + (2 * pp_enc_comp),
                pcmp.dysize + (2 * pp_enc_comp),
            ),
            layer=layer["pplus"],
        )
    )
    pplus.dxmin = pcmp.dxmin - pp_enc_comp
    pplus.dymin = pcmp.dymin - pp_enc_comp

    p_con = c.add_ref(
        via_stack(
            x_range=(pcmp.dxmin, pcmp.dxmax),
            y_range=(pcmp.dymin, pcmp.dymax),
            base_layer=layer["comp"],
            metal_level=1,
        )
    )  # pcmop contact

    # labels generation
    if label == 1:
        # n_label generation
        c.add_label(
            n_label,
            position=(
                n_con.dxmin + (n_con.dxsize / 2),
                n_con.dymin + (n_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

        # p_label generation
        c.add_label(
            p_label,
            position=(
                p_con.dxmin + (p_con.dxsize / 2),
                p_con.dymin + (p_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

    if volt == "5/6V":
        dg = c.add_ref(
            gf.components.rectangle(
                size=(
                    ncmp.dxmax - pcmp.dxmin + (2 * dg_enc_cmp),
                    ncmp.dysize + (2 * dg_enc_cmp),
                ),
                layer=layer["dualgate"],
            )
        )
        dg.dxmin = pcmp.dxmin - dg_enc_cmp
        dg.dymin = pcmp.dymin - dg_enc_cmp

    return c


@gf.cell
def diode_pw2dw(
    la: float = 0.1,
    wa: float = 0.1,
    cw: float = 0.1,
    volt: str = "3.3V",
    pcmpgr: bool = False,
    label: bool = False,
    p_label: str = "",
    n_label: str = "",
) -> gf.Component:
    """Used to draw LVPWELL/DNWELL diode by specifying parameters.

    Args:
        la: anode length.
        wa: anode width.
        cw: cathode width.
        volt: operating voltage of the diode [3.3V, 5V/6V]
        pcmpgr: if True, pcmpgr will be added.
        label: if True, labels will be added.
        p_label: p contact label.
        n_label: n contact label.

    """
    c = gf.Component()

    comp_spacing: float = 0.48
    np_enc_comp: float = 0.16
    pp_enc_comp: float = 0.16

    dg_enc_dn = 0.5

    lvpwell_enc_pcmp = 0.16
    dn_enc_lvpwell = 2.5

    con_size = 0.22
    con_sp = 0.28
    con_comp_enc = 0.07

    pcmpgr_enc_dn = 2.5

    # p generation
    pcmp = c.add_ref(gf.components.rectangle(size=(wa, la), layer=layer["comp"]))
    pplus = c.add_ref(
        gf.components.rectangle(
            size=(
                pcmp.dxsize + (2 * pp_enc_comp),
                pcmp.dysize + (2 * pp_enc_comp),
            ),
            layer=layer["pplus"],
        )
    )
    pplus.dxmin = pcmp.dxmin - pp_enc_comp
    pplus.dymin = pcmp.dymin - pp_enc_comp
    diode_mk = c.add_ref(
        gf.components.rectangle(
            size=(pcmp.dxsize, pcmp.dysize), layer=layer["diode_mk"]
        )
    )
    diode_mk.dxmin = pcmp.dxmin
    diode_mk.dymin = pcmp.dymin

    lvpwell = c.add_ref(
        gf.components.rectangle(
            size=(
                pcmp.dxsize + (2 * lvpwell_enc_pcmp),
                pcmp.dysize + (2 * lvpwell_enc_pcmp),
            ),
            layer=layer["lvpwell"],
        )
    )
    lvpwell.dxmin = pcmp.dxmin - lvpwell_enc_pcmp
    lvpwell.dymin = pcmp.dymin - lvpwell_enc_pcmp

    p_con = c.add_ref(
        via_stack(
            x_range=(pcmp.dxmin, pcmp.dxmax),
            y_range=(pcmp.dymin, pcmp.dymax),
            base_layer=layer["comp"],
            metal_level=1,
        )
    )  # pcomp_contact

    # n generation
    ncmp = c.add_ref(gf.components.rectangle(size=(cw, la), layer=layer["comp"]))
    ncmp.dxmax = pcmp.dxmin - comp_spacing
    nplus = c.add_ref(
        gf.components.rectangle(
            size=(
                ncmp.dxsize + (2 * np_enc_comp),
                ncmp.dysize + (2 * np_enc_comp),
            ),
            layer=layer["nplus"],
        )
    )
    nplus.dxmin = ncmp.dxmin - np_enc_comp
    nplus.dymin = ncmp.dymin - np_enc_comp

    n_con = c.add_ref(
        via_stack(
            x_range=(ncmp.dxmin, ncmp.dxmax),
            y_range=(ncmp.dymin, ncmp.dymax),
            base_layer=layer["comp"],
            metal_level=1,
        )
    )  # ncomp contact

    # labels generation
    if label == 1:
        # n_label generation
        c.add_label(
            n_label,
            position=(
                n_con.dxmin + (n_con.dxsize / 2),
                n_con.dymin + (n_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

        # p_label generation
        c.add_label(
            p_label,
            position=(
                p_con.dxmin + (p_con.dxsize / 2),
                p_con.dymin + (p_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

    dn_rect = c.add_ref(
        gf.components.rectangle(
            size=(
                lvpwell.dxsize + (2 * dn_enc_lvpwell),
                lvpwell.dysize + (2 * dn_enc_lvpwell),
            ),
            layer=layer["dnwell"],
        )
    )

    dn_rect.dxmin = lvpwell.dxmin - dn_enc_lvpwell
    dn_rect.dymin = lvpwell.dymin - dn_enc_lvpwell

    if pcmpgr == 1:
        c_temp_gr = gf.Component("temp_store guard ring")
        rect_pcmpgr_in = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(
                    (dn_rect.dxmax - dn_rect.dxmin) + 2 * pcmpgr_enc_dn,
                    (dn_rect.dymax - dn_rect.dymin) + 2 * pcmpgr_enc_dn,
                ),
                layer=layer["comp"],
            )
        )
        rect_pcmpgr_in.dmove(
            (dn_rect.dxmin - pcmpgr_enc_dn, dn_rect.dymin - pcmpgr_enc_dn)
        )
        rect_pcmpgr_out = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(
                    (rect_pcmpgr_in.dxmax - rect_pcmpgr_in.dxmin) + 2 * cw,
                    (rect_pcmpgr_in.dymax - rect_pcmpgr_in.dymin) + 2 * cw,
                ),
                layer=layer["comp"],
            )
        )
        rect_pcmpgr_out.dmove((rect_pcmpgr_in.dxmin - cw, rect_pcmpgr_in.dymin - cw))
        c.add_ref(
            gf.boolean(
                A=rect_pcmpgr_out,
                B=rect_pcmpgr_in,
                operation="A-B",
                layer=layer["comp"],
            )
        )  # guardring Bulk

        psdm_in = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(
                    (rect_pcmpgr_in.dxmax - rect_pcmpgr_in.dxmin) - 2 * pp_enc_comp,
                    (rect_pcmpgr_in.dymax - rect_pcmpgr_in.dymin) - 2 * pp_enc_comp,
                ),
                layer=layer["pplus"],
            )
        )
        psdm_in.dmove(
            (
                rect_pcmpgr_in.dxmin + pp_enc_comp,
                rect_pcmpgr_in.dymin + pp_enc_comp,
            )
        )
        psdm_out = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(
                    (rect_pcmpgr_out.dxmax - rect_pcmpgr_out.dxmin) + 2 * pp_enc_comp,
                    (rect_pcmpgr_out.dymax - rect_pcmpgr_out.dymin) + 2 * pp_enc_comp,
                ),
                layer=layer["pplus"],
            )
        )
        psdm_out.dmove(
            (
                rect_pcmpgr_out.dxmin - pp_enc_comp,
                rect_pcmpgr_out.dymin - pp_enc_comp,
            )
        )
        c.add_ref(
            gf.boolean(A=psdm_out, B=psdm_in, operation="A-B", layer=layer["pplus"])
        )  # guardring psdm

        # generatingg contacts

        c.add_ref(
            via_generator(
                x_range=(
                    rect_pcmpgr_in.dxmin + con_size,
                    rect_pcmpgr_in.dxmax - con_size,
                ),
                y_range=(rect_pcmpgr_out.dymin, rect_pcmpgr_in.dymin),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # bottom contact

        c.add_ref(
            via_generator(
                x_range=(
                    rect_pcmpgr_in.dxmin + con_size,
                    rect_pcmpgr_in.dxmax - con_size,
                ),
                y_range=(rect_pcmpgr_in.dymax, rect_pcmpgr_out.dymax),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # upper contact

        c.add_ref(
            via_generator(
                x_range=(rect_pcmpgr_out.dxmin, rect_pcmpgr_in.dxmin),
                y_range=(
                    rect_pcmpgr_in.dymin + con_size,
                    rect_pcmpgr_in.dymax - con_size,
                ),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # right contact

        c.add_ref(
            via_generator(
                x_range=(rect_pcmpgr_in.dxmax, rect_pcmpgr_out.dxmax),
                y_range=(
                    rect_pcmpgr_in.dymin + con_size,
                    rect_pcmpgr_in.dymax - con_size,
                ),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # left contact

        comp_m1_in = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(rect_pcmpgr_in.dxsize, rect_pcmpgr_in.dysize),
                layer=layer["metal1"],
            )
        )

        comp_m1_out = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(
                    (comp_m1_in.dxsize) + 2 * cw,
                    (comp_m1_in.dysize) + 2 * cw,
                ),
                layer=layer["metal1"],
            )
        )
        comp_m1_out.dmove((rect_pcmpgr_in.dxmin - cw, rect_pcmpgr_in.dymin - cw))
        c.add_ref(
            gf.boolean(
                A=rect_pcmpgr_out,
                B=rect_pcmpgr_in,
                operation="A-B",
                layer=layer["metal1"],
            )
        )  # guardring metal1

    if volt == "5/6V":
        dg = c.add_ref(
            gf.components.rectangle(
                size=(
                    dn_rect.dxsize + (2 * dg_enc_dn),
                    dn_rect.dysize + (2 * dg_enc_dn),
                ),
                layer=layer["dualgate"],
            )
        )
        dg.dxmin = dn_rect.dxmin - dg_enc_dn
        dg.dymin = dn_rect.dymin - dg_enc_dn

    return c


@gf.cell
def diode_dw2ps(
    la: float = 0.1,
    wa: float = 0.1,
    cw: float = 0.1,
    volt: str = "3.3V",
    pcmpgr: bool = False,
    label: bool = False,
    p_label: str = "",
    n_label: str = "",
) -> gf.Component:
    """Used to draw LVPWELL/DNWELL diode by specifying parameters.

    Args:
        la: anode length.
        wa: anode width.
        cw: cathode width.
        volt: operating voltage of the diode [3.3V, 5V/6V].
        pcmpgr: True if pwell guardring is required.
        label: True if labels are required.
        p_label: label for pwell.
        n_label: label for nwell.

    """
    c = gf.Component()

    if volt == "5/6V":
        dn_enc_ncmp = 0.66
    else:
        dn_enc_ncmp = 0.62

    comp_spacing = 0.32
    np_enc_comp: float = 0.16
    pp_enc_comp: float = 0.16

    con_size = 0.22
    con_sp = 0.28
    con_comp_enc = 0.07

    dg_enc_dn = 0.5

    pcmpgr_enc_dn = 2.5

    if (wa < ((2 * cw) + comp_spacing)) or (la < ((2 * cw) + comp_spacing)):
        ncmp = c.add_ref(gf.components.rectangle(size=(wa, la), layer=layer["comp"]))

        n_con = c.add_ref(
            via_stack(
                x_range=(ncmp.dxmin, ncmp.dxmax),
                y_range=(ncmp.dymin, ncmp.dymax),
                base_layer=layer["comp"],
                metal_level=1,
            )
        )  # ncomp_contact

        nplus = c.add_ref(
            gf.components.rectangle(
                size=(
                    ncmp.dxsize + (2 * np_enc_comp),
                    ncmp.dysize + (2 * np_enc_comp),
                ),
                layer=layer["nplus"],
            )
        )
        nplus.dxmin = ncmp.dxmin - np_enc_comp
        nplus.dymin = ncmp.dymin - np_enc_comp
    else:
        c_temp = gf.Component("temp_store guard ring")
        ncmp_in = c_temp.add_ref(
            gf.components.rectangle(
                size=(wa - (2 * cw), la - (2 * cw)),
                layer=layer["comp"],
            )
        )
        ncmp_out = c_temp.add_ref(
            gf.components.rectangle(
                size=(wa, la),
                layer=layer["comp"],
            )
        )
        ncmp_out.dmove((ncmp_in.dxmin - cw, ncmp_in.dymin - cw))
        ncmp = c.add_ref(
            gf.boolean(
                A=ncmp_out,
                B=ncmp_in,
                operation="A-B",
                layer=layer["comp"],
            )
        )

        pplus_in = c_temp.add_ref(
            gf.components.rectangle(
                size=(
                    (ncmp_in.dxmax - ncmp_in.dxmin) - 2 * pp_enc_comp,
                    (ncmp_in.dymax - ncmp_in.dymin) - 2 * pp_enc_comp,
                ),
                layer=layer["pplus"],
            )
        )
        pplus_in.dmove(
            (
                ncmp_in.dxmin + pp_enc_comp,
                ncmp_in.dymin + pp_enc_comp,
            )
        )
        pplus_out = c_temp.add_ref(
            gf.components.rectangle(
                size=(
                    (ncmp_out.dxmax - ncmp_out.dxmin) + 2 * pp_enc_comp,
                    (ncmp_out.dymax - ncmp_out.dymin) + 2 * pp_enc_comp,
                ),
                layer=layer["pplus"],
            )
        )
        pplus_out.dmove(
            (
                ncmp_out.dxmin - pp_enc_comp,
                ncmp_out.dymin - pp_enc_comp,
            )
        )
        c.add_ref(
            gf.boolean(A=pplus_out, B=pplus_in, operation="A-B", layer=layer["nplus"])
        )  # nplus

        # generatingg contacts

        c.add_ref(
            via_generator(
                x_range=(
                    ncmp_in.dxmin + con_size,
                    ncmp_in.dxmax - con_size,
                ),
                y_range=(ncmp_out.dymin, ncmp_in.dymin),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # bottom contact

        c.add_ref(
            via_generator(
                x_range=(
                    ncmp_in.dxmin + con_size,
                    ncmp_in.dxmax - con_size,
                ),
                y_range=(ncmp_in.dymax, ncmp_out.dymax),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # upper contact

        n_con = c.add_ref(
            via_generator(
                x_range=(ncmp_out.dxmin, ncmp_in.dxmin),
                y_range=(
                    ncmp_in.dymin + con_size,
                    ncmp_in.dymax - con_size,
                ),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # left contact

        c.add_ref(
            via_generator(
                x_range=(ncmp_in.dxmax, ncmp_out.dxmax),
                y_range=(
                    ncmp_in.dymin + con_size,
                    ncmp_in.dymax - con_size,
                ),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # right contact

        comp_m1_in = c_temp.add_ref(
            gf.components.rectangle(
                size=(ncmp_in.dxsize, ncmp_in.dysize),
                layer=layer["metal1"],
            )
        )

        comp_m1_out = c_temp.add_ref(
            gf.components.rectangle(
                size=(
                    (comp_m1_in.dxsize) + 2 * cw,
                    (comp_m1_in.dxsize) + 2 * cw,
                ),
                layer=layer["metal1"],
            )
        )
        comp_m1_out.dmove((ncmp_in.dxmin - cw, ncmp_in.dymin - cw))
        c.add_ref(
            gf.boolean(
                A=ncmp_out,
                B=ncmp_in,
                operation="A-B",
                layer=layer["metal1"],
            )
        )  # guardring metal1

    # labels generation
    if label == 1:
        # n_label generation
        c.add_label(
            n_label,
            position=(
                n_con.dxmin + (n_con.dxsize / 2),
                n_con.dymin + (n_con.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

    # generate dnwell

    dn_rect = c.add_ref(
        gf.components.rectangle(
            size=(
                ncmp.dxsize + (2 * dn_enc_ncmp),
                ncmp.dysize + (2 * dn_enc_ncmp),
            ),
            layer=layer["dnwell"],
        )
    )
    dn_rect.dxmin = ncmp.dxmin - dn_enc_ncmp
    dn_rect.dymin = ncmp.dymin - dn_enc_ncmp

    diode_mk = c.add_ref(
        gf.components.rectangle(
            size=(dn_rect.dxsize, dn_rect.dysize), layer=layer["diode_mk"]
        )
    )
    diode_mk.dxmin = dn_rect.dxmin
    diode_mk.dymin = dn_rect.dymin

    if pcmpgr == 1:
        c_temp_gr = gf.Component("temp_store guard ring")
        rect_pcmpgr_in = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(
                    (dn_rect.dxmax - dn_rect.dxmin) + 2 * pcmpgr_enc_dn,
                    (dn_rect.dymax - dn_rect.dymin) + 2 * pcmpgr_enc_dn,
                ),
                layer=layer["comp"],
            )
        )
        rect_pcmpgr_in.dmove(
            (dn_rect.dxmin - pcmpgr_enc_dn, dn_rect.dymin - pcmpgr_enc_dn)
        )
        rect_pcmpgr_out = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(
                    (rect_pcmpgr_in.dxmax - rect_pcmpgr_in.dxmin) + 2 * cw,
                    (rect_pcmpgr_in.dymax - rect_pcmpgr_in.dymin) + 2 * cw,
                ),
                layer=layer["comp"],
            )
        )
        rect_pcmpgr_out.dmove((rect_pcmpgr_in.dxmin - cw, rect_pcmpgr_in.dymin - cw))
        c.add_ref(
            gf.boolean(
                A=rect_pcmpgr_out,
                B=rect_pcmpgr_in,
                operation="A-B",
                layer=layer["comp"],
            )
        )  # guardring Bulk

        psdm_in = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(
                    (rect_pcmpgr_in.dxmax - rect_pcmpgr_in.dxmin) - 2 * pp_enc_comp,
                    (rect_pcmpgr_in.dymax - rect_pcmpgr_in.dymin) - 2 * pp_enc_comp,
                ),
                layer=layer["pplus"],
            )
        )
        psdm_in.dmove(
            (
                rect_pcmpgr_in.dxmin + pp_enc_comp,
                rect_pcmpgr_in.dymin + pp_enc_comp,
            )
        )
        psdm_out = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(
                    (rect_pcmpgr_out.dxmax - rect_pcmpgr_out.dxmin) + 2 * pp_enc_comp,
                    (rect_pcmpgr_out.dymax - rect_pcmpgr_out.dymin) + 2 * pp_enc_comp,
                ),
                layer=layer["pplus"],
            )
        )
        psdm_out.dmove(
            (
                rect_pcmpgr_out.dxmin - pp_enc_comp,
                rect_pcmpgr_out.dymin - pp_enc_comp,
            )
        )
        c.add_ref(
            gf.boolean(A=psdm_out, B=psdm_in, operation="A-B", layer=layer["pplus"])
        )  # psdm

        # generatingg contacts

        c.add_ref(
            via_generator(
                x_range=(
                    rect_pcmpgr_in.dxmin + con_size,
                    rect_pcmpgr_in.dxmax - con_size,
                ),
                y_range=(rect_pcmpgr_out.dymin, rect_pcmpgr_in.dymin),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # bottom contact

        c.add_ref(
            via_generator(
                x_range=(
                    rect_pcmpgr_in.dxmin + con_size,
                    rect_pcmpgr_in.dxmax - con_size,
                ),
                y_range=(rect_pcmpgr_in.dymax, rect_pcmpgr_out.dymax),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # upper contact

        p_con = c.add_ref(
            via_generator(
                x_range=(rect_pcmpgr_out.dxmin, rect_pcmpgr_in.dxmin),
                y_range=(
                    rect_pcmpgr_in.dymin + con_size,
                    rect_pcmpgr_in.dymax - con_size,
                ),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # left contact

        c.add_ref(
            via_generator(
                x_range=(rect_pcmpgr_in.dxmax, rect_pcmpgr_out.dxmax),
                y_range=(
                    rect_pcmpgr_in.dymin + con_size,
                    rect_pcmpgr_in.dymax - con_size,
                ),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # right contact

        # labels generation
        if label == 1:
            # n_label generation
            c.add_label(
                p_label,
                position=(
                    p_con.dxmin + (p_con.dxsize / 2),
                    p_con.dymin + (p_con.dysize / 2),
                ),
                layer=layer["metal1_label"],
            )

        comp_m1_in = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(rect_pcmpgr_in.dxsize, rect_pcmpgr_in.dysize),
                layer=layer["metal1"],
            )
        )

        comp_m1_out = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(
                    (rect_pcmpgr_in.dxmax - rect_pcmpgr_in.dxmin) + 2 * cw,
                    (rect_pcmpgr_in.dymax - rect_pcmpgr_in.dymin) + 2 * cw,
                ),
                layer=layer["metal1"],
            )
        )
        comp_m1_out.dmove((rect_pcmpgr_in.dxmin - cw, rect_pcmpgr_in.dymin - cw))
        c.add_ref(
            gf.boolean(
                A=rect_pcmpgr_out,
                B=rect_pcmpgr_in,
                operation="A-B",
                layer=layer["metal1"],
            )
        )  # guardring metal1

    # generate dualgate

    if volt == "5/6V":
        dg = c.add_ref(
            gf.components.rectangle(
                size=(
                    dn_rect.dxsize + (2 * dg_enc_dn),
                    dn_rect.dysize + (2 * dg_enc_dn),
                ),
                layer=layer["dualgate"],
            )
        )
        dg.dxmin = dn_rect.dxmin - dg_enc_dn
        dg.dymin = dn_rect.dymin - dg_enc_dn

    return c


@gf.cell
def sc_diode(
    la: float = 0.1,
    wa: float = 0.1,
    cw: float = 0.1,
    m: int = 1,
    pcmpgr: bool = False,
    label: bool = False,
    p_label: str = "",
    n_label: str = "",
) -> gf.Component:
    """Used to draw N+/LVPWELL diode (Outside DNWELL) by specifying parameters.

    Args:
     la         : Float of diff length (anode)
     wa         : Float of diff width (anode)
     m          : Integer of number of fingers
     pcmpgr     : Boolean of using P+ Guard Ring for Deep NWELL devices only

    """
    c = gf.Component("sc_diode_dev")

    sc_enc_comp = 0.16
    sc_comp_spacing = 0.28
    dn_enc_sc_an = 1.4
    np_enc_comp = 0.03
    m1_w = 0.23
    pcmpgr_enc_dn = 2.5
    pp_enc_comp: float = 0.16

    con_size = 0.22
    con_sp = 0.28
    con_comp_enc = 0.07

    # cathode draw
    @gf.cell
    def sc_cathode_strap(size: Float2 = (0.1, 0.1)) -> gf.Component:
        """Returns sc_diode cathode array element.

        Args :
            size : size of cathode array element
        """
        c = gf.Component()

        ncmp = c.add_ref(gf.components.rectangle(size=size, layer=layer["comp"]))

        nplus = c.add_ref(
            gf.components.rectangle(
                size=(
                    ncmp.dxsize + (2 * np_enc_comp),
                    ncmp.dysize + (2 * np_enc_comp),
                ),
                layer=layer["nplus"],
            )
        )
        nplus.dxmin = ncmp.dxmin - np_enc_comp
        nplus.dymin = ncmp.dymin - np_enc_comp

        c.add_ref(
            via_stack(
                x_range=(ncmp.dxmin, ncmp.dxmax),
                y_range=(ncmp.dymin, ncmp.dymax),
                base_layer=layer["comp"],
                metal_level=1,
            )
        )  # ncomp contact

        return c

    @gf.cell
    def sc_anode_strap(size: Float2 = (0.1, 0.1)) -> gf.Component:
        """Returns sc_diode anode array element.

        Args :
            size : size of anode array element
        """
        c = gf.Component()
        cmp = c.add_ref(gf.components.rectangle(size=size, layer=layer["comp"]))
        c.add_ref(
            via_stack(
                x_range=(cmp.dxmin, cmp.dxmax),
                y_range=(cmp.dymin, cmp.dymax),
                base_layer=layer["comp"],
                metal_level=1,
            )
        )  # comp contact
        return c

    sc_an = sc_anode_strap(size=(wa, la))
    sc_cath = sc_cathode_strap(size=(cw, la))

    sc_cathode = c.add_ref(
        component=sc_cath,
        rows=1,
        columns=(m + 1),
        column_pitch=cw + wa + (2 * sc_comp_spacing),
    )

    cath_m1_xmin = sc_cathode.dxmin
    cath_m1_ymin = sc_cathode.dymin
    cath_m1_xmax = sc_cathode.dxmax

    cath_m1_v = c.add_ref(
        component=gf.components.rectangle(
            size=(
                cath_m1_xmax - cath_m1_xmin,
                cath_m1_ymin - sc_cathode.dymin + m1_w,
            ),
            layer=layer["metal1"],
        ),
        rows=1,
        columns=(m + 1),
        column_pitch=(cw + wa + (2 * sc_comp_spacing)),
    )

    cath_m1_v.dxmin = cath_m1_xmin
    cath_m1_v.dymax = cath_m1_ymin
    cath_m1_h = c.add_ref(
        gf.components.rectangle(size=(cath_m1_v.dxsize, m1_w), layer=layer["metal1"])
    )
    cath_m1_h.dxmin = cath_m1_v.dxmin
    cath_m1_h.dymax = cath_m1_v.dymin

    # cathode label generation
    if label == 1:
        c.add_label(
            n_label,
            position=(
                cath_m1_h.dxmin + (cath_m1_h.dxsize / 2),
                cath_m1_h.dymin + (cath_m1_h.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

    sc_anode = c.add_ref(
        component=sc_an,
        rows=1,
        columns=m,
        column_pitch=(wa + cw + (2 * sc_comp_spacing)),
    )
    sc_anode.dxmin = sc_cathode.dxmin + (cw + sc_comp_spacing)
    an_m1_xmin = sc_anode.dxmin
    an_m1_ymin = sc_anode.dymin
    an_m1_xmax = sc_anode.dxmax
    an_m1_ymax = sc_anode.dymax

    if m > 1:
        an_m1_v = c.add_ref(
            component=gf.components.rectangle(
                size=(
                    an_m1_xmax - an_m1_xmin,
                    cath_m1_ymin - sc_an.dymin + m1_w,
                ),
                layer=layer["metal1"],
            ),
            rows=1,
            columns=m,
            spacing=((cw + wa + (2 * sc_comp_spacing)), 0),
        )

        an_m1_v.dxmin = an_m1_xmin
        an_m1_v.dymin = an_m1_ymax

        an_m1_h = c.add_ref(
            gf.components.rectangle(size=(an_m1_v.dxsize, m1_w), layer=layer["metal1"])
        )
        an_m1_h.dxmin = an_m1_v.dxmin
        an_m1_h.dymin = an_m1_v.dymax

        # anode label generation
        if label == 1:
            c.add_label(
                p_label,
                position=(
                    an_m1_h.dxmin + (an_m1_h.dxsize / 2),
                    an_m1_h.dymin + (an_m1_h.dysize / 2),
                ),
                layer=layer["metal1_label"],
            )

    else:
        # anode label generation
        if label == 1:
            c.add_label(
                p_label,
                position=(
                    an_m1_xmin + ((an_m1_xmax - an_m1_xmin) / 2),
                    an_m1_ymin + ((an_m1_ymax - an_m1_ymin) / 2),
                ),
                layer=layer["metal1_label"],
            )

    # diode_mk
    diode_mk = c.add_ref(
        gf.components.rectangle(
            size=(
                sc_cathode.dxsize + (2 * sc_enc_comp),
                sc_cathode.dysize + (2 * sc_enc_comp),
            ),
            layer=layer["schottky_diode"],
        )
    )
    diode_mk.dxmin = sc_cathode.dxmin - sc_enc_comp
    diode_mk.dymin = sc_cathode.dymin - sc_enc_comp

    # dnwell
    dn_rect = c.add_ref(
        gf.components.rectangle(
            size=(
                sc_anode.dxsize + (2 * dn_enc_sc_an),
                sc_anode.dysize + (2 * dn_enc_sc_an),
            ),
            layer=layer["dnwell"],
        )
    )
    dn_rect.dxmin = sc_anode.dxmin - dn_enc_sc_an
    dn_rect.dymin = sc_anode.dymin - dn_enc_sc_an

    if pcmpgr == 1:
        c_temp_gr = gf.Component("temp_store guard ring")
        rect_pcmpgr_in = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(
                    (dn_rect.dxmax - dn_rect.dxmin) + 2 * pcmpgr_enc_dn,
                    (dn_rect.dymax - dn_rect.dymin) + 2 * pcmpgr_enc_dn,
                ),
                layer=layer["comp"],
            )
        )
        rect_pcmpgr_in.dmove(
            (dn_rect.dxmin - pcmpgr_enc_dn, dn_rect.dymin - pcmpgr_enc_dn)
        )
        rect_pcmpgr_out = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(
                    (rect_pcmpgr_in.dxmax - rect_pcmpgr_in.dxmin) + 2 * cw,
                    (rect_pcmpgr_in.dymax - rect_pcmpgr_in.dymin) + 2 * cw,
                ),
                layer=layer["comp"],
            )
        )
        rect_pcmpgr_out.dmove((rect_pcmpgr_in.dxmin - cw, rect_pcmpgr_in.dymin - cw))
        c.add_ref(
            gf.boolean(
                A=rect_pcmpgr_out,
                B=rect_pcmpgr_in,
                operation="A-B",
                layer=layer["comp"],
            )
        )  # guardring Bulk

        psdm_in = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(
                    (rect_pcmpgr_in.dxmax - rect_pcmpgr_in.dxmin) - 2 * pp_enc_comp,
                    (rect_pcmpgr_in.dymax - rect_pcmpgr_in.dymin) - 2 * pp_enc_comp,
                ),
                layer=layer["pplus"],
            )
        )
        psdm_in.dmove(
            (
                rect_pcmpgr_in.dxmin + pp_enc_comp,
                rect_pcmpgr_in.dymin + pp_enc_comp,
            )
        )
        psdm_out = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(
                    (rect_pcmpgr_out.dxmax - rect_pcmpgr_out.dxmin) + 2 * pp_enc_comp,
                    (rect_pcmpgr_out.dymax - rect_pcmpgr_out.dymin) + 2 * pp_enc_comp,
                ),
                layer=layer["pplus"],
            )
        )
        psdm_out.dmove(
            (
                rect_pcmpgr_out.dxmin - pp_enc_comp,
                rect_pcmpgr_out.dymin - pp_enc_comp,
            )
        )
        c.add_ref(
            gf.boolean(A=psdm_out, B=psdm_in, operation="A-B", layer=layer["pplus"])
        )  # psdm

        # generatingg contacts
        c.add_ref(
            via_generator(
                x_range=(
                    rect_pcmpgr_in.dxmin + con_size,
                    rect_pcmpgr_in.dxmax - con_size,
                ),
                y_range=(rect_pcmpgr_out.dymin, rect_pcmpgr_in.dymin),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # bottom contact

        c.add_ref(
            via_generator(
                x_range=(
                    rect_pcmpgr_in.dxmin + con_size,
                    rect_pcmpgr_in.dxmax - con_size,
                ),
                y_range=(rect_pcmpgr_in.dymax, rect_pcmpgr_out.dymax),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # upper contact

        c.add_ref(
            via_generator(
                x_range=(rect_pcmpgr_out.dxmin, rect_pcmpgr_in.dxmin),
                y_range=(
                    rect_pcmpgr_in.dymin + con_size,
                    rect_pcmpgr_in.dymax - con_size,
                ),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # right contact

        c.add_ref(
            via_generator(
                x_range=(rect_pcmpgr_in.dxmax, rect_pcmpgr_out.dxmax),
                y_range=(
                    rect_pcmpgr_in.dymin + con_size,
                    rect_pcmpgr_in.dymax - con_size,
                ),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # left contact

        comp_m1_in = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(rect_pcmpgr_in.dxsize, rect_pcmpgr_in.dysize),
                layer=layer["metal1"],
            )
        )

        comp_m1_out = c_temp_gr.add_ref(
            gf.components.rectangle(
                size=(
                    (comp_m1_in.dxsize) + 2 * cw,
                    (comp_m1_in.dysize) + 2 * cw,
                ),
                layer=layer["metal1"],
            )
        )
        comp_m1_out.dmove((rect_pcmpgr_in.dxmin - cw, rect_pcmpgr_in.dymin - cw))
        c.add_ref(
            gf.boolean(
                A=rect_pcmpgr_out,
                B=rect_pcmpgr_in,
                operation="A-B",
                layer=layer["metal1"],
            )
        )  # guardring metal1

    # creating layout and cell in klayout
    return c


if __name__ == "__main__":
    c = sc_diode()
    # c = diode_pd2nw()
    c.show()
